/*-
 * ~~~~~~licensing~~~~~~
 * object-factory
 * ==========
 * Copyright (C) 2020 - 2023 EmDev LLC
 * ==========
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * ~~~~~~/licensing~~~~~~
 */
package ru.entaxy.platform.base.objects.factory;

import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import com.google.gson.Gson;
import com.google.gson.JsonArray;
import com.google.gson.JsonElement;
import com.google.gson.JsonObject;

import ru.entaxy.platform.base.objects.factory.EntaxyFactory.CONFIGURATION.DIRECTIVES;
import ru.entaxy.platform.base.objects.factory.EntaxyFactory.CONFIGURATION.DIRECTIVES.OVERRIDE_MODE;
import ru.entaxy.platform.base.objects.factory.EntaxyFactory.FieldInfo;
import ru.entaxy.platform.base.objects.factory.EntaxyFactory.OutputInfo;
import ru.entaxy.platform.base.objects.factory.Importer.ImportInfo;
import ru.entaxy.platform.base.support.CommonUtils;
import ru.entaxy.platform.base.support.JSONUtils;

public class EntaxyFactoryUtils {

	public static final String PROP_HIERARCHY = "hierarchy";
	
	public static Gson gson = new Gson();
	
	public static String getEffectiveJson(EntaxyFactory factory) {
		JsonObject result = new JsonObject();
		
		// factory
		JsonObject factoryData = new JsonObject();
		factoryData.addProperty(EntaxyFactory.CONFIGURATION.FACTORY.ID, factory.getId());
		factoryData.addProperty(EntaxyFactory.CONFIGURATION.FACTORY.TYPE, factory.getType());
		factoryData.addProperty(EntaxyFactory.CONFIGURATION.FACTORY.CATEGORY, factory.getCategory());
		factoryData.addProperty(EntaxyFactory.CONFIGURATION.FACTORY.LABEL, factory.getLabel());
		factoryData.addProperty(EntaxyFactory.CONFIGURATION.FACTORY.DESCRIPTION, factory.getDescription());
		result.add(EntaxyFactory.CONFIGURATION.FACTORY_SECTION_NAME, factoryData);
		
		// typed data
		String type = factory.getType();
		JsonElement typeData = gson.toJsonTree(factory.getTypeInfo());
		result.add(type, typeData);
		
		// fields
		JsonObject fieldsData = new JsonObject();
		for (FieldInfo fi: factory.getFields()) {
			fieldsData.add(fi.getName(), fi.getJsonOrigin());
		}
		result.add(EntaxyFactory.CONFIGURATION.FIELDS_SECTION_NAME, fieldsData);
		
		// outputs
		JsonObject outputsData = new JsonObject();
		for (OutputInfo oi: factory.getOutputs()) {
			JsonObject outputData = new JsonObject();
			outputData.addProperty(EntaxyFactory.CONFIGURATION.OUTPUTS.ATTRIBUTES.IS_DEFAULT, oi.isDefault());
			outputData.addProperty(EntaxyFactory.CONFIGURATION.OUTPUTS.ATTRIBUTES.GENERATOR, oi.getGenerator());
			outputData.add(EntaxyFactory.CONFIGURATION.OUTPUTS.ATTRIBUTES.SCOPES
					, gson.toJsonTree(oi.getScopes()));
			if (oi.getConfig() != null)
				outputData.add(EntaxyFactory.CONFIGURATION.OUTPUTS.ATTRIBUTES.CONFIG
						, gson.toJsonTree(oi.getConfig()));
			JsonObject outputFields = new JsonObject();
			for (FieldInfo fi: oi.getFields()) {
				outputFields.add(fi.getName(), fi.getJsonOrigin());
			}
			outputData.add(EntaxyFactory.CONFIGURATION.OUTPUTS.ATTRIBUTES.FIELDS, outputFields);
			outputsData.add(oi.getType(), outputData);
		}
		result.add(EntaxyFactory.CONFIGURATION.OUTPUTS_SECTION_NAME, outputsData);
		
		return result.toString();
	}
	
	public static JsonObject cleanJson(JsonObject origin) {
		JsonObject result = origin.deepCopy();
		
		cleanObject(result, EntaxyFactory.CONFIGURATION.DIRECTIVES.getAllDirectives());
		
		return result;
	}
	
	public static void cleanObject(JsonObject origin, Set<String> toRemove) {
		for (String name: toRemove)
			origin.remove(name);
		for (Entry<String, JsonElement> entry: origin.entrySet())
			if (entry.getValue().isJsonObject())
				cleanObject(entry.getValue().getAsJsonObject(), toRemove);
	}
	
	public static interface FactoryConfigurationStorage {
		JsonObject getConfiguration(String factoryId);
	}
	
	public static JsonObject calculateEffectiveJson(JsonObject origin, FactoryConfigurationStorage storage) {
		JsonObject originCopy = origin.deepCopy();
		JsonObject result = originCopy;
		JsonObject error = new JsonObject();
		JsonObject errorData = new JsonObject();
		JsonArray errorReqs = new JsonArray();
		
		error.add("#CALC_ERROR", errorData);
		errorData.add("#REQS", errorReqs);
		
		// process imports in originCopy
		
		Importer importer = new Importer(originCopy);
		List<ImportInfo> imports = importer.findImports();
		if (!imports.isEmpty()) {
			final Map<String, JsonObject> factoryJsonMap = new HashMap<>();
			imports.stream()
				.forEach(imp->
					imp.importDescriptors.stream()
						.filter(id->CommonUtils.isValid(id.factoryId))
						.forEach(id->factoryJsonMap.put(id.factoryId, null))
				);
			boolean inconsistent = false;
			for (String factoryId: factoryJsonMap.keySet()) {
				JsonObject factoryJson;
				if ("#".equals(factoryId))
					factoryJson = origin;
				else
					factoryJson = storage.getConfiguration(factoryId);
				if (factoryJson == null) {
					inconsistent = true;
					errorReqs.add(factoryId);
				} else {
					factoryJsonMap.put(factoryId, factoryJson);
				}
					
			}
			if (inconsistent)
				return error;
			
			importer.addImportedSources(factoryJsonMap);
			importer.processImports(imports);
			
		}
		
		// process inheritance
		String parent = getParentFromJson(origin);
		if (CommonUtils.isValid(parent)) {
			JsonObject parentObject = storage.getConfiguration(parent);
			if (parentObject == null) {
				errorReqs.add(parent);
				return error;
			}
			result = parentObject.deepCopy();
			prepareTypeInfo(result, originCopy);
			processObjectOverriding(result, originCopy, OVERRIDE_MODE.UPDATE);
		} else {
			result = originCopy;
		}
		
		return result;
	}

	public static void prepareTypeInfo(JsonObject parent, JsonObject child) {
		JsonElement parentTypeE = JSONUtils.findElement(parent, "factory.type");
		JsonElement childTypeE = JSONUtils.findElement(child, "factory.type");
		String parentType = "";
		String childType = "";
		if ((parentTypeE!=null) && parentTypeE.isJsonPrimitive())
			parentType = parentTypeE.getAsString();
		if ((childTypeE!=null) && childTypeE.isJsonPrimitive())
			childType = childTypeE.getAsString();
		
		JsonObject parentProperties = getTypeProperties(parent);
		JsonObject childProperties = getTypeProperties(child);
		if (childProperties==null)
			return;
		if (!parentType.equals(childType) && CommonUtils.isValid(childType)) {
			// copy parent type info with child type
			parent.add(childType, getTypeProperties(parent).deepCopy());
		}
		JsonElement parentHierarchy = JSONUtils.findElement(parentProperties, PROP_HIERARCHY);
		JsonArray childHierarchy = new JsonArray();
		if ((parentHierarchy != null) && parentHierarchy.isJsonArray()) {
			childHierarchy.addAll(parentHierarchy.getAsJsonArray());
		}
		childHierarchy.add(JSONUtils.findElement(parent, "factory.id"));
		childProperties.add(PROP_HIERARCHY, childHierarchy);
	}
	
	public static JsonObject resolveVariants(JsonObject root) {
		JsonObject result = root.deepCopy();
		processObjectVariants(result, root);
		return result;
	} 
	
	public static void processObjectVariants(JsonObject currentObject, JsonObject root) {
		JsonObject variants = getVariants(currentObject);
		if (variants != null) {
			String property = variants.has("property")
					?variants.get("property").getAsString()
					:"";
			if (CommonUtils.isValid(property)) {
				JsonObject values = (variants.has("values") && variants.get("values").isJsonObject())
						?variants.get("values").getAsJsonObject()
						:null;
				if (values != null) {
					JsonObject typeProperties = getTypeProperties(root);
					JsonElement je = JSONUtils.findElement(typeProperties, property);
					if (je != null) {
						String value = je.getAsString();
						if (values.has(value) && values.get(value).isJsonObject()) {
							JsonObject variant = values.get(value).getAsJsonObject();
							for (Entry<String, JsonElement> entry: variant.entrySet()) {
								currentObject.remove(entry.getKey());
								currentObject.add(entry.getKey(), entry.getValue().deepCopy());
							}
						}
					}
				}
			}
			currentObject.remove(DIRECTIVES.VARIANTS);
		}
		for (Entry<String, JsonElement> entry: currentObject.entrySet()) {
			if (entry.getValue().isJsonObject())
				processObjectVariants(entry.getValue().getAsJsonObject(), root);
		}
	}
	
	public static JsonObject getVariants(JsonObject currentObject) {
		if (currentObject.has(DIRECTIVES.VARIANTS) && currentObject.get(DIRECTIVES.VARIANTS).isJsonObject())
			return currentObject.get(DIRECTIVES.VARIANTS).getAsJsonObject();
		return null;
	}

	public static JsonObject getTypeProperties(JsonObject currentObject) {
		JsonElement type = JSONUtils.findElement(currentObject, "factory.type");
		if (type == null)
			return new JsonObject();
		String typeValue = type.getAsString();
		if (!CommonUtils.isValid(typeValue))
			return new JsonObject();
		JsonElement typeProperties = JSONUtils.findElement(currentObject, typeValue);
		if (typeProperties.isJsonObject())
			return typeProperties.getAsJsonObject();
		return new JsonObject();
	}
	
	public static String getParentFromJson(JsonObject jsonObject) {
		if (!jsonObject.has(EntaxyFactory.CONFIGURATION.FACTORY_SECTION_NAME))
			return null;
		JsonObject fs = jsonObject.get(EntaxyFactory.CONFIGURATION.FACTORY_SECTION_NAME).getAsJsonObject();
		if (fs.has(EntaxyFactory.CONFIGURATION.FACTORY.PARENT))
			return fs.get(EntaxyFactory.CONFIGURATION.FACTORY.PARENT).getAsString();
		return null;
	}
	
	public static void processObjectOverriding(JsonObject currentObject, JsonObject newObject, OVERRIDE_MODE defaultMode) {
		
		OVERRIDE_MODE mode = getOverrideMode(currentObject, defaultMode);		
		
		if (OVERRIDE_MODE.IGNORE.equals(mode))
			return;
		
		if (OVERRIDE_MODE.REPLACE.equals(mode)) {
			Set<String> set = new HashSet<String>(currentObject.keySet());
			for (String entry: set)
				currentObject.remove(entry);
			for (Entry<String, JsonElement> entry: newObject.entrySet())
				currentObject.add(entry.getKey(), entry.getValue().deepCopy());
			return;
		}
		
		if (OVERRIDE_MODE.APPEND.equals(mode)) {
			for (Entry<String, JsonElement> entry: newObject.entrySet())
				if (!currentObject.has(entry.getKey()))
					currentObject.add(entry.getKey(), entry.getValue().deepCopy());
			return;
		}

		if (OVERRIDE_MODE.UPDATE.equals(mode)) {
			
			
			// update existing
			Set<String> set = new HashSet<String>(currentObject.keySet());
			for (String key: set) {
				if (newObject.has(key)) {
					JsonElement currentElement = currentObject.get(key);
					JsonElement newElement = newObject.get(key);
					if (currentElement.isJsonObject() && newElement.isJsonObject()) {
						processObjectOverriding(currentElement.getAsJsonObject(), newElement.getAsJsonObject(), mode);
					} else {
						currentObject.remove(key);
						currentObject.add(key, newElement.deepCopy());
					}
				}
			}
			
			// add new
			for (Entry<String, JsonElement> entry: newObject.entrySet())
				if (!currentObject.has(entry.getKey()))
					currentObject.add(entry.getKey(), entry.getValue().deepCopy());
			
		}
		
	}
	
	public static EntaxyFactory.CONFIGURATION.DIRECTIVES.OVERRIDE_MODE getOverrideMode(JsonObject object, OVERRIDE_MODE defaultMode) {
		OVERRIDE_MODE result = defaultMode;
		if (object.has(EntaxyFactory.CONFIGURATION.DIRECTIVES.OVERRIDE)) {
			try {
				OVERRIDE_MODE mode = OVERRIDE_MODE.valueOfLabel(object.get(EntaxyFactory.CONFIGURATION.DIRECTIVES.OVERRIDE).getAsString());
				if (mode != null)
					result = mode;
			} catch (Exception e) {
				
			}
		}
		return result;
	}
	
}
