/*-
 * ~~~~~~licensing~~~~~~
 * resources-service
 * ==========
 * Copyright (C) 2020 - 2023 EmDev LLC
 * ==========
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * ~~~~~~/licensing~~~~~~
 */
package ru.entaxy.esb.resources.provider;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.net.MalformedURLException;
import java.net.URL;
import java.nio.ByteBuffer;
import java.nio.channels.FileChannel;
import java.nio.channels.FileLock;
import java.nio.file.Path;
import java.nio.file.Paths;

import org.apache.commons.io.IOUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import ru.entaxy.esb.resources.EntaxyResource;
import ru.entaxy.esb.resources.EntaxyResourceProvider;

public class FileResourceProvider implements EntaxyResourceProvider {

	private static final Logger log = LoggerFactory.getLogger(FileResourceProvider.class);
	
	private String protocol;
	private String rootDirectory;
	private File root;
	
	@Override
	public String getProtocol() {
		return protocol;
	}
	
	public void setProtocol(String protocol) {
		this.protocol = protocol;
	}
	
	public String getRootDirectory() {
		return rootDirectory;
	}

	public void setRootDirectory(String rootDirectory) {
		this.rootDirectory = rootDirectory;
		root = new File(this.rootDirectory);
		if (!root.exists())
			root.mkdirs();
	}

	@Override
	public EntaxyResource getResource(String location) {
		Path rootPath = root.toPath();
		Path targetPath = Paths.get(rootPath.toString(), location);
		log.debug("TARGET PATH :: " + targetPath.toAbsolutePath().toString());
		File targetFile = new File(targetPath.toAbsolutePath().toString());
		log.debug("TARGET FILE :: " + targetFile.getAbsolutePath());
		return new FileResource(targetFile);
	}
	
	public static class FileResource implements EntaxyResource {
		
		private File file;
		
		public FileResource(File file) {
			this.file = file;
		}

		@Override
		public InputStream getInputStream() {
			if (!exists())
				return null;
			try {
				return file.toURI().toURL().openStream();
			} catch (IOException e) {
				FileResourceProvider.log.error("Error getting resource", e);
				return null;
			}
		}

		@Override
		public String getAsString() {
			if (!exists())
				return null;
			try {
				return IOUtils.toString(getInputStream());
			} catch (IOException e) {
				FileResourceProvider.log.error("Error getting resource as string", e);
			}
			return null;
		}

		@Override
		public void save(InputStream inputStream) {
			if (!file.exists())
				try {
					log.debug("\nMAKE DIRS", file.getParentFile().getAbsolutePath());
					file.getParentFile().mkdirs();
					// file.createNewFile();
				} catch (Exception e1) {
					log.error("Error creating resource", e1);
				}
			try (FileOutputStream fileOutputStream = new FileOutputStream(file)){
			     FileChannel channel = fileOutputStream.getChannel();
			     FileLock lock = channel.lock();
			     byte[] bytes = IOUtils.toByteArray(inputStream);
			     channel.write(ByteBuffer.wrap(bytes));
			     
			     lock.release();
			     channel.close();
			     IOUtils.closeQuietly(fileOutputStream);
			     
			} catch (Exception e) {
				log.error("Error saving resource", e);
			}
		}

		@Override
		public boolean exists() {
			return file.exists();
		}

		@Override
		public URL getLocation() {
			try {
				return this.file.toURI().toURL();
			} catch (MalformedURLException e) {
				return null;
			}
		}
		
	}
	
}
