/*-
 * ~~~~~~licensing~~~~~~
 * test-producers
 * ==========
 * Copyright (C) 2020 - 2023 EmDev LLC
 * ==========
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * ~~~~~~/licensing~~~~~~
 */
package ru.entaxy.platform.core.producer.executor.generationmodel;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;

import ru.entaxy.platform.core.producer.executor.objectmodel.FactoredObject;
import ru.entaxy.platform.core.producer.executor.objectmodel.FactoredObjectProxy;
import ru.entaxy.platform.core.producer.executor.objectmodel.FactoredObjectRef;
import ru.entaxy.platform.core.producer.executor.objectmodel.ObjectModel;

public class GenerationModel {

	protected ObjectModel objectModel;
	
	protected List<AbstractTask> tasks = new LinkedList<>();
	
	public Map<Object, List<AbstractTask>> objectToTasksMap = new HashMap<>();
	
	public GenerationModel(ObjectModel objectModel) {
		this.objectModel = objectModel;
	}
	
	public void load() {
		for (FactoredObject fo: objectModel.objects)
			if (!fo.isEmbedded())
				tasks.addAll(generateTasksForObject(fo));
	}
	
	protected List<AbstractTask> generateTasksForObject(FactoredObject fo) {

		if (fo instanceof FactoredObjectProxy)
			return Collections.singletonList(new ResolveObjectTask((FactoredObjectProxy)fo));
		
		List<AbstractTask> localPreTasks = new LinkedList<>();
		List<AbstractTask> localPostTasks = new LinkedList<>();

		// first process embedded objects 

		List<FactoredObject> embedded = fo.embedded.stream().filter(f->(f instanceof FactoredObject))
				.map(f -> (FactoredObject)f)
				.collect(Collectors.toList());
		for (FactoredObject embeddedObject: embedded) {
			List<AbstractTask> tasksForEmbedded = generateTasksForObject(embeddedObject);
			localPreTasks.addAll(tasksForEmbedded);
		}

		// process links to externalized embedded objects 

		List<FactoredObjectRef> embeddedLinks = fo.embedded.stream().filter(f->(f instanceof FactoredObjectRef))
				.map(f -> (FactoredObjectRef)f)
				.filter(f -> !f.isBackRef())
				.filter(f -> f.isLink())
				.collect(Collectors.toList());
		for (FactoredObjectRef embeddedRef: embeddedLinks) {
			FactoredObject embeddedObject = embeddedRef.getTargetObject();
			List<AbstractTask> tasksForEmbedded = generateTasksForObject(embeddedObject);
			
			// check the dependencies:
			// if embedded depends on main
			// then it must be generates after the main
			
			List<FactoredObjectRef> refs = objectModel.refs.stream()
					.filter(r -> r.isDefinedIn(embeddedObject))
					.filter(r -> fo.equals(r.getTargetObject()))
					.collect(Collectors.toList());
			if (refs.isEmpty())
				localPreTasks.addAll(tasksForEmbedded);
			else
				localPostTasks.addAll(tasksForEmbedded);
		}

		// process embedded refs 
		
		List<FactoredObjectRef> embeddedRefs = fo.embedded.stream().filter(f->(f instanceof FactoredObjectRef))
				.map(f -> (FactoredObjectRef)f)
				.filter(f -> !f.isBackRef())
				.filter(f -> !f.isLink() || f.isResolveAlways())
				.collect(Collectors.toList());
		for (FactoredObjectRef embeddedRef: embeddedRefs) {
			
			if (!embeddedRef.isLink() || embeddedRef.isResolveAlways()) {
				localPreTasks.add(new ResolveRefTask(embeddedRef));
				if (!embeddedRef.isLink())
					if (!embeddedRef.isRefByValueOnly())
						localPreTasks.add(new GenerateRefTask(embeddedRef));
			}
		}
		
		List<AbstractTask> currentTasks = new ArrayList<>(localPreTasks);
		
		if (CalculateTask.hasCalculations(fo))
			currentTasks.add(new CalculateTask(fo));
		
		GenerateTask generateTask = new GenerateTask(fo); 

		currentTasks.add(generateTask);

		for (AbstractTask at: localPreTasks)
			if (((at instanceof GenerateTask) || (at instanceof GenerateRefTask)) && !at.isToMerge() )
				currentTasks.add(new MergeTask(at, generateTask));
		
		currentTasks.addAll(localPostTasks);
		
		for (AbstractTask at: localPostTasks)
			if (((at instanceof GenerateTask) || (at instanceof GenerateRefTask)) && !at.isToMerge())
				currentTasks.add(new MergeTask(at, generateTask));
	
		return currentTasks;
	}

	public List<AbstractTask> getTasks() {
		return tasks;
	}
	
}
