/*-
 * ~~~~~~licensing~~~~~~
 * object-producer-core
 * ==========
 * Copyright (C) 2020 - 2023 EmDev LLC
 * ==========
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * ~~~~~~/licensing~~~~~~
 */
package ru.entaxy.platform.core.producer.executor.objectmodel;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.google.gson.JsonElement;
import com.google.gson.JsonObject;

import ru.entaxy.platform.base.objects.factory.EntaxyFactory;
import ru.entaxy.platform.base.support.CommonUtils;
import ru.entaxy.platform.base.support.JSONUtils;
import ru.entaxy.platform.base.support.osgi.OSGIUtils;
import ru.entaxy.platform.core.producer.api.EntaxyProducerService;

public class Calculation extends AbstractItem {

	private static final Logger log = LoggerFactory.getLogger(Calculation.class);
	
	// TODO move const to EntaxyObject
	public static final String FIELD_IS_CALCULATED = "isCalculated";
	
	public static final String OWNER_PREFIX = "#OWNER";
	public static final String OWNER_OF_TYPE_PREFIX = "#OWNER_OF_TYPE";
	public static final String FACTORY_PREFIX = "#FACTORY";
	public static final String TYPE_DATA_PREFIX = "#TYPEINFO";
	
	public static final String OWNER_TOKEN = "#OWNER#";
	public static final String FACTORY_TOKEN = "#FACTORY#";
	public static final String TYPE_DATA_TOKEN = "#TYPEINFO#";
	
	public static final Object NOT_ENOUGH_DATA = new Object();
	
	public static boolean isCalculated(JsonObject jsonObject) {
		boolean result = false;
		
		// old-style calculated
		if (jsonObject.has(FIELD_IS_CALCULATED))
			try {
				result = jsonObject.get(FIELD_IS_CALCULATED).getAsBoolean();
			} catch (Exception e) {
				// return false;
			}
		
		// new-style calculated
		if (jsonObject.has(EntaxyFactory.CONFIGURATION.DIRECTIVES.CALCULATED)) {
			result = jsonObject.get(EntaxyFactory.CONFIGURATION.DIRECTIVES.CALCULATED).isJsonObject();
		}
		return result;
	}
	
	protected ExpressionCalculation expressionCalculation;
	
	protected String targetPath;
	protected boolean removeResolved = false;
	
	protected boolean lazy = true;
	
	public Calculation(ObjectModel objectModel) {
		super(objectModel);
	}
	
	@Override
	public void wrap(JsonObject object, Map<String, Object> context, String path) {
		
		JsonObject content = object;
		if (object.has(EntaxyFactory.CONFIGURATION.DIRECTIVES.CALCULATED))
			content = object.get(EntaxyFactory.CONFIGURATION.DIRECTIVES.CALCULATED).getAsJsonObject();

		if (!ExpressionCalculation.isExpressionCalculation(content))
			return;
		
		super.wrap(object, context, path);
		
		
		this.expressionCalculation = new ExpressionCalculation();
		this.expressionCalculation.init(content);
		
		AbstractFactored contextOwner = AbstractFactored.getContextOwner(context);
		if (contextOwner instanceof FactoredObject)
			this.definedIn = (FactoredObject)contextOwner;
		
		contextOwner.addInternal(this);
		
		computeRelativePath(contextOwner.jsonPath);
		
		if (content.has("targetPath")) {
			JsonElement tp = content.get("targetPath");
			if (tp.isJsonNull()) {
				this.targetPath = this.relativePath;
			} else {
				this.targetPath = tp.getAsString();
			}
		} else {
			this.targetPath = this.relativePath;
		}
		
		if (content.has("removeResolved"))
			try {
				this.removeResolved = content.get("removeResolved").getAsBoolean();
			} catch (Exception e) {
				// NOOP
			}

		if (content.has("lazy"))
			try {
				this.lazy = content.get("lazy").getAsBoolean();
			} catch (Exception e) {
				// NOOP
			}

		
		// System.out.println("\n\n===>>>> [CALCULATION] CREATED [" + contextOwner.factoryId + "/" + this.relativePath + "]");
		
	}
	public Object calculate(ObjectModel objectModel) {
		if (this.definedIn == null)
			return null;
		return calculate(objectModel, this.definedIn);
	} 

	public Object calculate(ObjectModel objectModel, FactoredObject factoredObject) {
		return this.expressionCalculation.calculate(objectModel, factoredObject);
	}
	
	public String getTargetPath() {
		return targetPath;
	}

	public boolean isRemoveResolved() {
		return removeResolved;
	}

	public boolean isLazy() {
		return lazy;
	}

	public static class ExpressionCalculation {
		
		// TODO move const to EntaxyObject
		public static final String ATTR_EXPRESSION = "expression";
		
		public static final String ATTR_RESULT_TYPE = "resultType";
		
		public static final String ATTR_ALLOW_OBJECTS = "allowObjects";
		public static final String ATTR_ALLOW_ARRAYS = "allowArrays"; 
		public static final String ATTR_ALLOW_PRIMITIVES = "allowPrimitives";
		public static final String ATTR_ALLOW_NULLS = "allowNulls"; 
		
		protected static final Object TOKEN_NOT_FOUND = new Object();
		protected static final Object RESULT_NOT_ALLOWED = new Object();
		
		protected static final RESULT_TYPE DEFAULT_RESULT_TYPE = RESULT_TYPE.STRING;
		
		public enum RESULT_TYPE {
			STRING,
			LIST,
			LIST_OR_SINGLE,
			MAP;
		}
		
		JsonObject jsonObject;
		
		String expression;
		
		String resultTypeValue = "String";
		
		boolean allowObjects = true;
		boolean allowArrays = true;
		boolean allowPrimitives = true;
		boolean allowNulls = false;
		
		List<String> toResolve = new ArrayList<>();
		Map<String, Object> resolved = new HashMap<>();
		
		protected RESULT_TYPE resultType = DEFAULT_RESULT_TYPE;
		
		public static boolean isExpressionCalculation(JsonObject jsonObject) {
			return jsonObject.has(ATTR_EXPRESSION);
		}
		
		public void init(JsonObject jsonObject) {
			this.jsonObject = jsonObject;
			this.expression = jsonObject.get(ATTR_EXPRESSION).getAsString();
			if (jsonObject.has(ATTR_RESULT_TYPE)) {
				this.resultTypeValue = jsonObject.get(ATTR_RESULT_TYPE).getAsString();
				if (CommonUtils.isValid(resultTypeValue)) {
					if (RESULT_TYPE.valueOf(this.resultTypeValue.toUpperCase()) != null)
						this.resultType = RESULT_TYPE.valueOf(this.resultTypeValue.toUpperCase());
				}
			}
			try {
				allowObjects = jsonObject.get(ATTR_ALLOW_OBJECTS).getAsBoolean();
			} catch (Exception e) {
				// NOOP
			}
			try {
				allowArrays = jsonObject.get(ATTR_ALLOW_ARRAYS).getAsBoolean();
			} catch (Exception e) {
				// NOOP
			}
			try {
				allowPrimitives = jsonObject.get(ATTR_ALLOW_PRIMITIVES).getAsBoolean();
			} catch (Exception e) {
				// NOOP
			}
			try {
				allowNulls = jsonObject.get(ATTR_ALLOW_NULLS).getAsBoolean();
			} catch (Exception e) {
				// NOOP
			}
		    Pattern pattern = Pattern.compile("\\$\\{.+?\\}");
		    Matcher matcher = pattern.matcher(this.expression);
		    while (matcher.find()) {
		        String value = this.expression.substring(matcher.start(), matcher.end());
		        // System.out.println("\\nFOUND :: " + value);
		        this.toResolve.add(value.substring(2, value.length()-1));
		    }
		}
		
		public Object calculate(ObjectModel objectModel, FactoredObject factoredObject) {
			for (String expr: toResolve) {
				if (!resolved.containsKey(expr)) {
					Object result = resolve(expr, objectModel, factoredObject);
					if (!TOKEN_NOT_FOUND.equals(result) && !RESULT_NOT_ALLOWED.equals(result)) {
						resolved.put(expr, result);
					}
				}
			}
			toResolve.removeAll(resolved.keySet());
			if (toResolve.isEmpty()) {
				return doCalculate();
			}
			return NOT_ENOUGH_DATA;
		}
		
		public Object doCalculate() {
			String expr = expression;
			
			Object result = null;
			List<Object> list = new ArrayList<>();
			
			switch (this.resultType) {
			case LIST:
				list.addAll(resolved.values());
				result = list;
				break;
			case LIST_OR_SINGLE:
				list.addAll(resolved.values());
				result = list;
				if (list.size()==1)
					result = list.get(0);
				break;
			case MAP:
				// TODO
				break;
			default:
				for (Entry<String, Object> entry: resolved.entrySet()) {
					Object value = entry.getValue();
					String val = value==null?"":value.toString();
					expr = expr.replace("${" + entry.getKey() + "}", val);
				}
				result = expr;
				break;
			}
			
			return result;
		}
		
		protected boolean startsWithOwnerPrefix(String token) {
			if (!CommonUtils.isValid(token))
				return false;
			return token.startsWith(OWNER_PREFIX)
					|| token.startsWith(OWNER_OF_TYPE_PREFIX);
		}
		
		public Object resolve(String token, ObjectModel objectModel, FactoredObject factoredObject) {

			FactoredObject currentObject = factoredObject;
			String currentToken = token;
			
			while (startsWithOwnerPrefix(currentToken) && currentObject!=null) {
				int ind = currentToken.substring(1).indexOf("#.");
				String subtoken;
				if (ind < 0) {
					subtoken = currentToken;
					currentToken = "";
				} else {
					subtoken = currentToken.substring(0, ind+2);
					currentToken = currentToken.substring(ind+3);
				}
				if (Calculation.OWNER_TOKEN.equals(subtoken)) {
					currentObject = currentObject.getOwnerObject();
				} else 
				if (subtoken.startsWith(Calculation.OWNER_OF_TYPE_PREFIX)) {
					int i1 = subtoken.indexOf(':');
					if (i1<0) {
						log.warn("Incorrect expression: [{}]", subtoken);
						break;
					}
					String typeToFind = subtoken.substring(i1+1, subtoken.length()-1);
					
					while (currentObject != null) {
						currentObject = currentObject.getOwnerObject();
						if (currentObject != null)
							if (typeToFind.equals(currentObject.getObjectType()))
								break;
					}
					
				} else break;	
			}
			if (currentObject == null)
				return TOKEN_NOT_FOUND;
			if (!CommonUtils.isValid(currentToken))
				return TOKEN_NOT_FOUND;
			JsonObject jsonObject = currentObject.origin; 
			if (currentToken.startsWith(FACTORY_TOKEN)) {
				String factoryId = currentObject.factoryId;
				try {
					EntaxyProducerService entaxyProducerService = OSGIUtils.services()
							.ofClass(EntaxyProducerService.class).get();
					EntaxyFactory f = entaxyProducerService.findFactoryById(factoryId);
					jsonObject = JSONUtils.getJsonRootObject(f.getJsonConfiguration());
					currentToken = currentToken.substring(Calculation.FACTORY_TOKEN.length());
					if (currentToken.startsWith("."))
						currentToken = currentToken.substring(1);
					
					if (currentToken.startsWith(TYPE_DATA_TOKEN)) {
						JsonElement je = jsonObject.get(f.getType());
						if ((je != null) && je.isJsonObject() )
							jsonObject = je.getAsJsonObject();
						else
							jsonObject = null;
						
						currentToken = currentToken.substring(Calculation.TYPE_DATA_TOKEN.length());
						if (currentToken.startsWith("."))
							currentToken = currentToken.substring(1);
					}
					
				} catch (Exception e) {
					return TOKEN_NOT_FOUND;
				}
			}
			JsonElement result = null;
			if (jsonObject != null)
				if (CommonUtils.isValid(currentToken))
					result = findElement(jsonObject, currentToken);
				else
					result = jsonObject;
			
			if (result == null)
				return TOKEN_NOT_FOUND;
			if (!valueIsAllowed(result))
				return RESULT_NOT_ALLOWED;
			
			
			return JSONUtils.element2object(result);
		}
		
		protected boolean valueIsAllowed(JsonElement value) {
			if (value.isJsonNull() && !allowNulls)
				return false;
			if (value.isJsonObject())
				if (!allowObjects)
					return false;
				else {
					if (value.getAsJsonObject().has(EntaxyFactory.CONFIGURATION.DIRECTIVES.CALCULATED))
						return false;
				}
			if (value.isJsonArray() && !allowArrays)
				return false;
			if (value.isJsonPrimitive() && !allowPrimitives)
				return false;
			return true;
		}
		
		public JsonElement findElement(JsonObject jsonObject, String pathFragment) {
			JsonElement result = null;
			if (pathFragment.startsWith("?.")) {
				result = findElement(jsonObject, pathFragment.substring(2), "");
			} else {
				result = getElement(jsonObject, pathFragment);
			}
			
			return result;
		}
		public JsonElement getElement(JsonObject jsonObject, String pathFragment) {
			int ind = pathFragment.indexOf(".");
			String currentFragment;
			String newFragment;
			if (ind > 0) {
				currentFragment = pathFragment.substring(0, ind);
				newFragment = pathFragment.substring(ind+1);
			} else {
				currentFragment = pathFragment;
				newFragment = null;
			}
			if (!jsonObject.has(currentFragment))
				return null;
			JsonElement je = jsonObject.get(currentFragment);
			if (newFragment == null)
				return je;
			if (!je.isJsonObject())
				return null;
			return getElement(je.getAsJsonObject(), newFragment);
		}
		
		public JsonElement findElement(JsonObject jsonObject, String pathFragment, String parentPath) {
			
			for (Entry<String, JsonElement> entry: jsonObject.entrySet()) {
				
				String currentPath = parentPath + "." + entry.getKey();
				if (currentPath.endsWith(pathFragment))
					return entry.getValue();
				
			}
			
			for (Entry<String, JsonElement> entry: jsonObject.entrySet()) {
				
				if (!entry.getValue().isJsonObject())
					continue;
				
				String currentPath = parentPath + "." + entry.getKey();
				
				JsonElement currentResult = findElement(entry.getValue().getAsJsonObject(), pathFragment, currentPath);
				
				if (currentResult != null)
					return currentResult;
			}
			
			
			return null;
		}
		
	}
	
}
