/*-
 * ~~~~~~licensing~~~~~~
 * object-producing-resources-support
 * ==========
 * Copyright (C) 2020 - 2023 EmDev LLC
 * ==========
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * ~~~~~~/licensing~~~~~~
 */
package ru.entaxy.platform.core.producer.resources;

import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import org.osgi.service.component.ComponentContext;
import org.osgi.service.component.annotations.Activate;
import org.osgi.service.component.annotations.Component;
import org.osgi.service.component.annotations.Reference;
import org.osgi.service.component.annotations.ReferenceCardinality;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.google.gson.Gson;
import com.google.gson.JsonElement;
import com.google.gson.JsonObject;

import ru.entaxy.esb.platform.runtime.base.connecting.generator.Generated;
import ru.entaxy.esb.resources.EntaxyResourceService;
import ru.entaxy.platform.base.objects.EntaxyObject;
import ru.entaxy.platform.base.objects.EntaxyObject.FIELDS;
import ru.entaxy.platform.base.objects.factory.EntaxyFactory;
import ru.entaxy.platform.base.objects.factory.EntaxyFactory.CONFIGURATION.FACTORY;
import ru.entaxy.platform.base.objects.factory.EntaxyFactory.CONFIGURATION.OUTPUTS;
import ru.entaxy.platform.base.objects.factory.EntaxyFactory.FieldInfo;
import ru.entaxy.platform.base.support.CommonUtils;
import ru.entaxy.platform.base.support.JSONUtils;
import ru.entaxy.platform.core.producer.api.EntaxyProducerService;
import ru.entaxy.platform.core.producer.api.ExecutionPlan.ExecutionPlanUpdate;
import ru.entaxy.platform.core.producer.api.ProducerResult;
import ru.entaxy.platform.core.producer.api.ProducerResult.CommandResult;
import ru.entaxy.platform.core.producer.api.ProducingCommandExecutor;
import ru.entaxy.platform.core.producer.executor.AbstractCommandExecutor;
import ru.entaxy.platform.core.producer.executor.CommandExecutor;
import ru.entaxy.platform.core.producer.executor.objectmodel.Calculation;
import ru.entaxy.platform.core.producer.executor.objectmodel.FactoredObject;
import ru.entaxy.platform.core.producer.executor.objectmodel.FactoredObjectProxy;
import ru.entaxy.platform.core.producer.executor.objectmodel.ObjectModel;

@Component(service = ProducingCommandExecutor.class, immediate = true)
@CommandExecutor(id = "process-resources", predecessors = {"enrich"}, descendants = {"validate"})
public class ProcessResourcesCommand extends AbstractCommandExecutor implements ProducingCommandExecutor {

	private static final Logger log = LoggerFactory.getLogger(ProcessResourcesCommand.class);
	
	@Reference(cardinality = ReferenceCardinality.MANDATORY)
	EntaxyProducerService entaxyProducerServiceLocal;

	@Reference(cardinality = ReferenceCardinality.MANDATORY)
	EntaxyResourceService entaxyResourceService;
	
	@Reference(cardinality = ReferenceCardinality.MANDATORY)
	ResourceDataProcessor resourceDataProcessor;
	
	public ProcessResourcesCommand() {
		super(null);
	}
	
	protected ProcessResourcesCommand(EntaxyProducerService entaxyProducerService) {
		super(entaxyProducerService);
	}
	
	@Activate
	public void activate(ComponentContext componentContext) {
		this.entaxyProducerService = this.entaxyProducerServiceLocal;
		this.entaxyProducerService.registerCommand(this);
	}
	
	@Override
	protected boolean doExecute(ProducerResult currentResult, CommandResult commandResult,
			Map<String, Object> instructions) throws Exception {

		ObjectModel objectModel = currentResult.findResultObject(ObjectModel.class);
		objectModel.startTracking();

		JsonObject incomingJson = objectModel.getJsonCurrent().deepCopy();

		for (FactoredObject fo: objectModel.objects) {
			
			// skip proxies
			if (fo instanceof FactoredObjectProxy)
				continue;

			String factoryId = fo.factoryId;
			
			EntaxyFactory factory = entaxyProducerService.findFactoryById(factoryId);

			JsonObject objectOrigin = fo.origin;
			
			
			if (JSONUtils.setValue(objectOrigin, FIELDS.PROPERTIES, new JsonObject(), true)) {
				printOutput("[DIRTY] CREATED 'properties' for [" + fo.getObjectId() + "/" + fo.getObjectType() + "]");
				objectModel.setDirty();
			}

			JsonObject objectProperties = objectOrigin.get(FIELDS.PROPERTIES).getAsJsonObject();
			
			String outputType = CommonUtils.isValid(fo.getOutputType())
					?fo.getOutputType()
					:factory.getDefaultOutput()!=null
						?factory.getDefaultOutput().getType()
						:OUTPUTS.OUTPUT_TYPE_INIT;
			
			List<FieldInfo> factoryFields = factory.getFields(outputType); 
			
			log.debug("\n FIELDS FOR " + factory.getId() + "/" + outputType + ":" + factoryFields.size());
			
			// if the value is JsonObject
			// and represents some resource
			// we first resolve it to actual value
			JsonObject newProperties = new JsonObject();
			for (Entry<String, JsonElement> entry: objectProperties.entrySet()) {
				
				if (entry.getValue().isJsonObject())
					if (entry.getValue().getAsJsonObject().has(ResourceDataProcessor.ENTAXY_RESOURCE_KEY)) {
						JsonElement jsonDescriptor = entry.getValue().getAsJsonObject().get(ResourceDataProcessor.ENTAXY_RESOURCE_KEY);
						ResourceDescriptor resourceDescriptor = ResourceDescriptor.fromJson(jsonDescriptor);
						if (resourceDescriptor == null)
							continue;
						Object newData = resourceDataProcessor.getResourceDataFromDescriptor(resourceDescriptor);
						JsonElement newValue = (new Gson()).toJsonTree(newData);
						
						newProperties.add(entry.getKey(), newValue);
					}
			}
			for (Entry<String, JsonElement> entry: newProperties.entrySet()) {
				if (JSONUtils.setValue(objectProperties, entry.getKey(), entry.getValue().deepCopy(), false)) {
					printOutput("[DIRTY] UPDATED " + entry.getKey() + " for [" + fo.getObjectId() + "/" + fo.getObjectType() + "]");
					objectModel.setDirty();
				}
				
			}
			
			
			// for the field defined as a resource
			// we must make an object 
			for (FieldInfo fi: factoryFields) {
				log.debug("\n--> FIELD :: " + fi.getName());
				log.debug("\n" + objectProperties.toString());
				JsonObject jsonFi = fi.getJsonOrigin();
				if (!objectProperties.has(fi.getName()))
					continue;
				log.debug("\nORIGIN:\n" + fi.getJsonOrigin().toString());
				if (jsonFi.has(ResourceDataProcessor.ENTAXY_RESOURCE_KEY)) {
					log.debug("\n--> FIELD :: " + fi.getName() + " IS A RESOURCE");
					ResourceDescriptor descriptor = ResourceDescriptor.fromJson(jsonFi.get(ResourceDataProcessor.ENTAXY_RESOURCE_KEY));
					if (descriptor == null)
						continue;
					JsonElement currentValue = objectProperties.get(fi.getName());
					if (currentValue.isJsonObject()) {
						JsonObject currentObject = currentValue.getAsJsonObject();
						if (currentObject.has("isRef")) {
							continue;
						}
					}
					JsonElement newValueElement = jsonFi.get(ResourceDataProcessor.ENTAXY_RESOURCE_KEY);
					if (!newValueElement.isJsonObject())
						continue;
					JsonObject properties = newValueElement.getAsJsonObject().deepCopy();
					
					// add default location if not set
					if (!properties.has(ResourceDataProcessor.PROP_LOCATION)) {
						JsonObject locationData = new JsonObject();
						locationData.addProperty(Calculation.FIELD_IS_CALCULATED, true);
						locationData.addProperty("lazy", true);
						locationData.addProperty("expression"
								, "object-resources/${#OWNER#.objectId}/" + fi.getName() + "." + fi.getType().split(":")[0]);
						properties.add(ResourceDataProcessor.PROP_LOCATION, locationData);
					}
					
					properties.add("content", currentValue);
					properties.addProperty("type", fi.getType());
					JsonObject newValue = new JsonObject();
					newValue.addProperty(FACTORY.TYPE, "entaxy.resource");
					newValue.addProperty(EntaxyObject.FIELDS.FACTORY_ID, "entaxy-resource"); 
					
					JsonObject refConfig = new JsonObject();
					refConfig.addProperty(EntaxyObject.FIELDS.REF_FIELD, "content");
					refConfig.addProperty("generated.targetType", Generated.GENERATED_TYPE_BLUEPRINT_FRAGMENT);
					newValue.add("refConfig", refConfig);
					
					// newValue.addProperty("__skip", true);
					newValue.add(FIELDS.PROPERTIES, properties);
					if (jsonFi.has("config") && jsonFi.get("config").isJsonObject()) {
						JsonObject configObject = jsonFi.get("config").getAsJsonObject();
						for (Entry<String, JsonElement> entry: configObject.entrySet()) {
							if (!newValue.has(entry.getKey()))
								newValue.add(entry.getKey(), entry.getValue().deepCopy());
						}
					}
					JSONUtils.setValue(objectProperties, fi.getName(), newValue, false);
					objectModel.setDirty();
				}
			}
			
		}
		
		if (objectModel.stopTracking()) {

			// remove ##embedded
			for (FactoredObject fo: objectModel.objects)
				fo.origin.remove(FactoredObject.EMBEDDED_FIELD);

			
			commandResult.planUpdate = ExecutionPlanUpdate.create()
//					.updateInstructions().target("enrich").value("skip", true).complete()
					.reset().target("analyze").complete();
		}
		
		JsonObject outgoingJson = objectModel.getJsonCurrent();
		
		printOutput("\n== INCOMING JSON ==\n");
		printOutput(incomingJson.toString());
		printOutput("\n== OUTGOING JSON ==\n");
		printOutput(outgoingJson.toString());
		
		commandResult.resultObject(outgoingJson);

		return true;
	}

}
