/*-
 * ~~~~~~licensing~~~~~~
 * object-producing-resources-support
 * ==========
 * Copyright (C) 2020 - 2023 EmDev LLC
 * ==========
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * ~~~~~~/licensing~~~~~~
 */
package ru.entaxy.platform.core.producer.resources.processors;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;

import org.osgi.service.component.annotations.Component;
import org.osgi.service.component.annotations.Reference;
import org.osgi.service.component.annotations.ReferenceCardinality;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import ru.entaxy.esb.resources.EntaxyResource;
import ru.entaxy.esb.resources.EntaxyResourceService;
import ru.entaxy.platform.base.support.CommonUtils;
import ru.entaxy.platform.base.support.xml.CommonXMLUtils;
import ru.entaxy.platform.core.producer.resources.EntaxyResourceProcessor;

@Component (service = EntaxyResourceProcessor.class, immediate = true)
public class MergeResourceProcessor implements EntaxyResourceProcessor {

	private static final Logger log = LoggerFactory.getLogger(MergeResourceProcessor.class);

	public static final String PROP_SOURCE = "source";
	public static final String PROP_SOURCE_ELEMENT = "sourceElement";
	public static final String PROP_TARGET_ELEMENT = "targetElement";
	
	@Reference(cardinality = ReferenceCardinality.MANDATORY)
	EntaxyResourceService entaxyResourceService;
	
	@Override
	public String getName() {
		return "merge";
	}

	@Override
	public Object process(Object value, Map<String, Object> properties) {
		if (value == null)
			return value;
		
		Object valueToReturn = null;
		
		String sourceLocation = properties.getOrDefault(PROP_SOURCE, "").toString();
		if (!CommonUtils.isValid(sourceLocation)) {
			log.warn("'source' not defined in [{}]", properties);
			return value;
		}
		
		EntaxyResource resource = entaxyResourceService.getResource(sourceLocation);
		if (!resource.exists()) {
			log.warn("Source [{}] not found, was defined in [{}]", sourceLocation, properties);
			return value;
		}
		
		String sourceElement = properties.getOrDefault(PROP_SOURCE_ELEMENT, "").toString();
		if (!CommonUtils.isValid(sourceElement)) {
			log.warn("Source element not defined in [{}]", properties);
			return value;
		}

		String targetElement = properties.getOrDefault(PROP_TARGET_ELEMENT, "").toString();
		if (!CommonUtils.isValid(targetElement)) {
			log.warn("Source element not defined in [{}]", properties);
			return value;
		}
		
		String targetXMLString = value.toString();
		String sourceXMLString = resource.getAsString();
		
		try {
			Document targetDocument = CommonXMLUtils.parseString(false, targetXMLString);
			Document sourceDocument = CommonXMLUtils.parseString(false, sourceXMLString);
			
			NodeList nodesToReplace = targetDocument.getElementsByTagName(targetElement);
			if (nodesToReplace.getLength() == 0)
				return value;
			
			NodeList replacements = sourceDocument.getElementsByTagName(sourceElement);
			if (replacements.getLength() == 0)
				return value;
			
			Node replacement = replacements.item(0);
			NodeList replacementChildren = replacement.getChildNodes();
			
			List<Node> toRemove = new ArrayList<>();
			
			for (int i=0; i<nodesToReplace.getLength(); i++) {
				Node toReplace = nodesToReplace.item(i);
				for (int j=0; j<replacementChildren.getLength(); j++) {
					Node adopted = targetDocument.adoptNode(replacementChildren.item(j).cloneNode(true));
					toReplace.getParentNode().insertBefore(adopted, toReplace);
				}
				toRemove.add(toReplace);
			}
			
			for (Node n: toRemove)
				n.getParentNode().removeChild(n);
			
			valueToReturn = CommonXMLUtils.doc2string(targetDocument);
			
		} catch (Exception e) {
			log.error("Error merging documents", e);
			return value;
		}
		
		return valueToReturn;
	}

}
