/*-
 * ~~~~~~licensing~~~~~~
 * object-producing-resources-support
 * ==========
 * Copyright (C) 2020 - 2023 EmDev LLC
 * ==========
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * ~~~~~~/licensing~~~~~~
 */
package ru.entaxy.platform.core.producer.resources.processors;

import java.io.StringReader;
import java.io.StringWriter;
import java.util.Map;

import javax.xml.transform.ErrorListener;
import javax.xml.transform.Source;
import javax.xml.transform.TransformerException;
import javax.xml.transform.stream.StreamSource;

import org.osgi.service.component.annotations.Component;
import org.osgi.service.component.annotations.Reference;
import org.osgi.service.component.annotations.ReferenceCardinality;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import net.sf.saxon.s9api.Processor;
import net.sf.saxon.s9api.QName;
import net.sf.saxon.s9api.SaxonApiException;
import net.sf.saxon.s9api.Serializer;
import net.sf.saxon.s9api.XdmAtomicValue;
import net.sf.saxon.s9api.XsltCompiler;
import net.sf.saxon.s9api.XsltExecutable;
import net.sf.saxon.s9api.XsltTransformer;
import ru.entaxy.esb.resources.EntaxyResource;
import ru.entaxy.esb.resources.EntaxyResourceService;
import ru.entaxy.platform.base.support.CommonUtils;
import ru.entaxy.platform.core.producer.resources.EntaxyResourceProcessor;

@Component (service = EntaxyResourceProcessor.class, immediate = true)
public class XSLTSaxonResourceProcessor implements EntaxyResourceProcessor {

	public static final String PROP_SOURCE = "source";
	public static final String PROP_PARAMETERS = "parameters";
	
	private static final Logger log = LoggerFactory.getLogger(XSLTSaxonResourceProcessor.class);

	@Reference(cardinality = ReferenceCardinality.MANDATORY)
	EntaxyResourceService entaxyResourceService;

	@Override
	public String getName() {
		return "saxon";
	}

	@Override
	public Object process(Object value, Map<String, Object> properties) {

		log.info("SAXON :: processing [{}]", value);
		
		if (value == null)
			return value;
		
		String xsltLocation = properties.getOrDefault(PROP_SOURCE, "").toString();
		if (!CommonUtils.isValid(xsltLocation)) {
			log.warn("XSL not defined in [{}]", properties);
			return value;
		}
		
		EntaxyResource resource = entaxyResourceService.getResource(xsltLocation);
		if (!resource.exists()) {
			log.warn("XSL [{}] not found, was defined in [{}]", xsltLocation, properties);
			return value;
		}
		
		
		Processor processor = new Processor(false);
		XsltCompiler compiler = processor.newXsltCompiler();
		XsltExecutable stylesheet;
		StringWriter result = new StringWriter();
		try {
			compiler.setErrorListener(new ErrorListener() {
				
				@Override
				public void warning(TransformerException exception) throws TransformerException {
					log.warn("SAXON COMPILER:", exception);
				}
				
				@Override
				public void fatalError(TransformerException exception) throws TransformerException {
					log.error("SAXON COMPILER:", exception);
				}
				
				@Override
				public void error(TransformerException exception) throws TransformerException {
					log.error("SAXON COMPILER:", exception);
				}
			});
			stylesheet = compiler.compile(new StreamSource(resource.getInputStream()));
			Serializer out = processor.newSerializer(result);
			out.setOutputProperty(Serializer.Property.METHOD, "xml");
			out.setOutputProperty(Serializer.Property.INDENT, "yes");
			XsltTransformer transformer = stylesheet.load();
			Source xml = new StreamSource(new StringReader(value.toString()));
			transformer.setSource(xml);
			transformer.setDestination(out);
	        if (properties.containsKey(PROP_PARAMETERS)) {
	        	Object obj = properties.get(PROP_PARAMETERS);
	        	if (obj instanceof Map) {
	        		Map<String, Object> map = (Map<String, Object>)obj;
	        		for (Map.Entry<String, Object> entry: map.entrySet()) {
	        			transformer.setParameter(
	        					new QName(entry.getKey())
	        					, new XdmAtomicValue(entry.getValue().toString()));
	        		}
	        	}
	        }
			transformer.transform();		
		} catch (SaxonApiException e) {
			log.error("Error transforming [" + value.toString() + "] with [" + xsltLocation + "]", e);
			return value;
		}
		return result.toString();
	}

}
