/*-
 * ~~~~~~licensing~~~~~~
 * test-producers
 * ==========
 * Copyright (C) 2020 - 2023 EmDev LLC
 * ==========
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * ~~~~~~/licensing~~~~~~
 */
package ru.entaxy.platform.base.objects.profile.producing;

import java.util.HashMap;
import java.util.Map;

import org.osgi.service.component.annotations.CollectionType;
import org.osgi.service.component.annotations.Component;
import org.osgi.service.component.annotations.Reference;
import org.osgi.service.component.annotations.ReferenceCardinality;
import org.osgi.service.component.annotations.ReferencePolicy;
import org.osgi.service.component.annotations.ReferencePolicyOption;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import ru.entaxy.esb.platform.runtime.base.connecting.generator.Generated;
import ru.entaxy.platform.base.support.CommonUtils;
import ru.entaxy.platform.base.objects.EntaxyObject.OBJECT_TYPES;
import ru.entaxy.platform.base.objects.factory.EntaxyFactory;
import ru.entaxy.platform.base.objects.factory.EntaxyFactoryException;
import ru.entaxy.platform.core.artifact.ArtifactCoordinates;
import ru.entaxy.platform.core.producer.api.EntaxyProducer;
import ru.entaxy.platform.core.producer.api.EntaxyProducerService;
import ru.entaxy.platform.core.producer.api.EntaxyWrappedFactory;
import ru.entaxy.platform.core.producer.impl.CommonObjectProducer;
import ru.entaxy.platform.core.producer.impl.EntaxyProducerInfo;
import ru.entaxy.platform.core.producer.wrapper.AbstractFactoryWrapper;
import ru.entaxy.platform.core.producer.wrapper.AbstractFactoryWrapper.GenerationProcessor;

@Component(service = {EntaxyProducer.class}, immediate = true)
@EntaxyProducerInfo(supportedTypes = {OBJECT_TYPES.PROFILE, OBJECT_TYPES.DEFAULT_ROUTE})
public class ProfileProducer extends CommonObjectProducer implements EntaxyProducer, GenerationProcessor {
	
	private static final Logger log = LoggerFactory.getLogger(ProfileProducer.class);
	
	public static final String PROPERTY_OBJECT_ID = "objectId";
	public static final String PROPERTY_SYSTEM_NAME = "systemName";
	public static final String TEMP_PREFIX = "TEMP-";
	
	protected Map<String, Map<String, String>> publishedFields = new HashMap<>(); 
	
	@SuppressWarnings("serial")
	public ProfileProducer() {
		super();
		publishedFields.put(OBJECT_TYPES.PROFILE, new HashMap<>() {{
			put("name", "systemName");
			put("system", "systemName");
		}});
		publishedFields.put(OBJECT_TYPES.DEFAULT_ROUTE, new HashMap<>() {{
			put("name", "systemName");
			put("system", "systemName");
		}});
	}
	
	@Reference(bind = "addFactory", unbind = "removeFactory", cardinality = ReferenceCardinality.MULTIPLE
			, collectionType = CollectionType.SERVICE, policy = ReferencePolicy.DYNAMIC, policyOption = ReferencePolicyOption.GREEDY)
	public void addFactory(EntaxyFactory factory) {
		AbstractFactoryWrapper wrapper = super.doAddFactory(factory);
		if (wrapper != null)
			wrapper.setGenerationProcessor(this);
	}
	
	// WE MUST DECLARE IT for @Reference annotation processor
	@Override
	public void removeFactory(EntaxyFactory factory) {
		super.removeFactory(factory);
	}
	
	@Override
	public void preGenerate(EntaxyWrappedFactory wrappedFactory, String outputType, String scope, String objectType, Map<String, Object> parameters)
			throws EntaxyFactoryException {
		log.debug("PreGenerate");
		
		if (OBJECT_TYPES.PROFILE.equals(objectType)) {
			// create objectId if it's not set
			String systemName = parameters.getOrDefault(PROPERTY_SYSTEM_NAME, "").toString();
			String objectId = parameters.getOrDefault(PROPERTY_OBJECT_ID, "").toString();
			log.debug("systemName:[{}], objectId:[{}]", systemName, objectId);
			if (!CommonUtils.isValid(objectId) || objectId.startsWith(TEMP_PREFIX))
				if (CommonUtils.isValid(systemName)) {
					parameters.put(PROPERTY_OBJECT_ID, systemName);
				}
		} else if (OBJECT_TYPES.DEFAULT_ROUTE.equals(objectType)) {
			// create objectId if it's not set
			String systemName = parameters.getOrDefault(PROPERTY_SYSTEM_NAME, "").toString();
			String objectId = parameters.getOrDefault(PROPERTY_OBJECT_ID, "").toString();
			log.debug("systemName:[{}], objectId:[{}]", systemName, objectId);
			if (!CommonUtils.isValid(objectId) || objectId.startsWith(TEMP_PREFIX))
				if (CommonUtils.isValid(systemName)) {
					parameters.put(PROPERTY_OBJECT_ID, systemName);
				}
		}
		
	}

	@Override
	public void postGenerate(EntaxyWrappedFactory wrappedFactory, String outputType, String scope, String objectType, Map<String, Object> parameters, Generated generated)
			throws EntaxyFactoryException {
		super.postGenerate(wrappedFactory, outputType, scope, objectType, parameters, generated);
		generated.getProperties().put(
				EntaxyProducerService.INSTRUCTIONS.ARTIFACT.VERSION_POLICY, 
				ArtifactCoordinates.VERSION_POLICY_DATED_EMBEDDED);
	}

	@Override
	protected Map<String, String> getFieldsToPublish(String outputType, String scope, String objectType,
			String objectId) {
		
		Map<String, String> result = super.getFieldsToPublish(outputType, scope, objectType, objectId);
		result.putAll(publishedFields.get(objectType));
		
		return result;
				
	}
	
}
