/*-
 * ~~~~~~licensing~~~~~~
 * entaxy-management-plugin
 * ==========
 * Copyright (C) 2020 - 2023 EmDev LLC
 * ==========
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * ~~~~~~/licensing~~~~~~
 */
var Entaxy;
(function (Entaxy) {
    Entaxy._module
    .component('entaxyProperties', {
        template:
        `
          <h2>
            {{$ctrl.title}}
          </h2>
            <div class="properties-container">
              <entaxy-modal-group-fields groups="$ctrl.groups" fields="$ctrl.formFields"
                    errors="$ctrl.errors" mode="$ctrl.mode" ng-if="$ctrl.groups && $ctrl.formFields"></entaxy-modal-group-fields>
            </div>
            <button type="submit" class="btn btn-primary" ng-click="$ctrl.save($ctrl.formFields)">{{$ctrl.submitBtnTitle}}</button>
        `,
        controller: entaxyPropertiesController
    })
    .name;

    function entaxyPropertiesController(workspace, operationsService, entaxyService) {
        'ngInject';

        var ctrl = this;

        ctrl.$onInit = function() {

            ctrl.itemName = 'Profile';
            ctrl.mode = Entaxy.MODAL_MODES.EDIT;

            let selectedMbean = workspace.getSelectedMBean();
            ctrl.title = 'Properties of ' + selectedMbean.title;
            ctrl.submitBtnTitle = Entaxy.getButtonTitleByMode(ctrl.mode);

            let profilesMbean = workspace.tree.findDescendant(child => child.objectName ? child.objectName.endsWith('category=profiles') : false);
            let mbeanName = profilesMbean.objectName;

            operationsService
            .executeOperation(mbeanName, { name: 'get' + ctrl.itemName + 'Config' }, [ selectedMbean.title ])
            .then(result => {
                ctrl.itemInfo = JSON.parse(result);
                let properties = Object.entries(ctrl.itemInfo.properties);
                ctrl.properties = properties.map(property => { return { name: property[0], value: property[1] }; });

                let factory = workspace.tree.findDescendant(child => child.title === ctrl.itemInfo.factoryId);
                ctrl.factories = [{ name: factory.title, mbeanName: factory.objectName }];

                makeFormFieldsAndGroups();
            }).catch(error => {
                Core.notification('danger', error, 5000);
                Entaxy.log.error(error);
            });
        }

        function makeFormFieldsAndGroups() {

            ctrl.formFields = [];

            ctrl.formFields.push({
                label: 'Factory Id',
                name: 'factoryId',
                type: 'java.lang.String',
                helpText: null,
                value: ctrl.factories[0].name,
                readOnly: true,
                required: true,
                group: 'general'
            });

            let mbeanName = ctrl.factories[0].mbeanName;

            operationsService.executeOperation(mbeanName, { name: 'getFields' }, [ 'init' ])
            .then((response) => {

                let groups = new Set();
                groups.add('general');

                _.forEach(JSON.parse(response), (field) => {

                    let formField = entaxyService.makeFormField(field, ctrl.itemInfo['objectId'], ctrl.properties, ctrl.mode);

                    ctrl.formFields.push(formField);
                    groups.add(formField.group);
                });

                ctrl.formFields.forEach(formField => formField.type = Entaxy.convertToHtmlInputType(formField.type));

                ctrl.groups = Array.from(groups).map((group) => { return { name: group, displayName: group }; });
            });
        }

        ctrl.save = function(fieldsOrigin) {

            let fields = JSON.parse(JSON.stringify(fieldsOrigin));

            entaxyService.validateFields(fields, ctrl.factories[0], ctrl.itemName)
            .then(errors => {
                ctrl.errors = errors;

                if (Object.keys(ctrl.errors).length === 0) {
                    let args = entaxyService.getArguments(fields, ctrl.factories);
                    update(args);
                }
            });
        }

        function update(args) {

            let objectId = args.fields.find((field) => field.name === 'objectId').value;

            let fields = objectId ? args.fields.filter((field) => field.name !== 'objectId') : args.fields;

            let mbeanName = args.factoryId.mbeanName;
            let properties = fields.reduce((obj, cur) => ({ ...obj, [cur.name] : cur.value }), {});
            let instructions = {"@LIFECYCLE":["general"]};

            operationsService.executeOperation(mbeanName, { name: 'createObjectByInstructions' }, [ objectId, 'public', instructions, properties ] )
            .then(result => {
                let tableHtml = Entaxy.createTableFromResponse(result);
                Entaxy.notification('success', tableHtml, 10000);
            }).catch(error => {
                Core.notification('danger', error, 5000);
                Entaxy.log.error(error);
            });
        }

    }
    entaxyPropertiesController.$inject = ['workspace', 'operationsService', 'entaxyService'];

})(Entaxy || (Entaxy = {}));
