/*-
 * ~~~~~~licensing~~~~~~
 * entaxy-management-plugin
 * ==========
 * Copyright (C) 2020 - 2023 EmDev LLC
 * ==========
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * ~~~~~~/licensing~~~~~~
 */
var Entaxy;
(function (Entaxy) {
    Entaxy._module
    .component('entaxyTable', {
        bindings: {
            itemName: '@',
            folderName: '@',
            statusName: '@',
            objectName: '@',
            isParentDependent: '<',
            extraPrimaryActionsConfig: '<'
        },
        template:
        `
          <pf-toolbar class="entaxy-toolbar" config="$ctrl.toolbarConfig"></pf-toolbar>
          <pf-table-view config="$ctrl.tableConfig"
                          columns="$ctrl.tableColumns"
                          action-buttons="$ctrl.tableActionButtons"
                          menu-actions="$ctrl.menuActions"
                          page-config="$ctrl.pageConfig"
                          items="$ctrl.viewedItems">
          </pf-table-view>
        `,
        controller: EntaxyTableController
    })
    .name;

    function EntaxyTableController(workspace, $location, $scope, $q, jolokia, jolokiaService, $uibModal, operationsService) {
        'ngInject';

        var ctrl = this;
        ctrl.workspace = workspace;
        var entaxyJmxDomain = localStorage['entaxyJmxDomain'] || "ru.entaxy.esb";

        let filterValues = Entaxy.getAllBundleStates();

        ctrl.tableConfig = {
            selectionMatchProp: 'name',
            showCheckboxes: false
        };

        ctrl.toolbarConfig = {
            filterConfig: {
                fields: [
                    {
                        id: 'name',
                        title: 'Name',
                        placeholder: 'Filter by Name...',
                        filterType: 'text'
                    },
                    {
                        id: 'status',
                        title: 'Status',
                        placeholder: 'Filter by Status...',
                        filterType: 'select',
                        filterValues: filterValues
                    }
                ],
                appliedFilters: [],
                onFilterChange: filterChange
            },
            isTableView: true
        };

        ctrl.tableColumns = [
            { header: 'Name', itemField: 'name' },
            { header: 'Status', itemField: 'status' }
        ];

        ctrl.pageConfig = {
            pageNumber: 1,
            pageSize: 10,
            pageSizeIncrements: [5, 10, 20, 50, 100]
        };

        function filterChange(filters) {
            applyFilters(filters);
            ctrl.toolbarConfig.filterConfig.resultsCount = ctrl.viewedItems.length;
        };

        function applyFilters(filters) {
            ctrl.viewedItems = [];
            if (filters && filters.length > 0) {
                ctrl.items.forEach(function (item) {
                    if (matchesFilters(item, filters)) {
                        ctrl.viewedItems.push(item);
                    }
                });
            } else {
                ctrl.viewedItems = ctrl.items;
            }
        };

        function matchesFilters(item, filters) {
            var matches = true;
            filters.forEach(function(filter) {
                if (!matchesFilter(item, filter)) {
                    matches = false;
                    return false;
                }
            });
            return matches;
        };

        function matchesFilter(item, filter) {
            var match = true;
            if (filter.id === 'name') {
                match = item.name.match(filter.value) !== null;
            } else if (filter.id === 'status') {
                match = item.status === filter.value;
            }
            return match;
        };

        ctrl.items = [];
        ctrl.viewedItems = [];

        ctrl.$onInit = function() {
            populateTable();

            let primaryActions = [ { name: 'Add ' + ctrl.itemName, actionFn: () => ctrl.showModal(Entaxy.MODAL_MODES.ADD) } ];

            if (ctrl.extraPrimaryActionsConfig) {
                let extraPrimaryActions = ctrl.extraPrimaryActionsConfig.primaryActions;
                primaryActions = primaryActions.concat(extraPrimaryActions);

                if (ctrl.extraPrimaryActionsConfig.isTableUpdateNeeded) {
                    ctrl.extraPrimaryActionsConfig.updateTableFn = populateTable;
                }
            }

            ctrl.toolbarConfig.actionsConfig = {
                primaryActions: primaryActions
            };

            ctrl.tableActionButtons = [
                { name: 'Start', title: 'Start ' + ctrl.itemName, actionFn: startItem },
                { name: 'Stop', title: 'Stop ' + ctrl.itemName, actionFn: stopItem },
                { name: 'Uninstall', title: 'Uninstall ' + ctrl.itemName, actionFn: uninstallItem }
            ];

            ctrl.menuActions = [
                { name: 'View Properties', title: 'View ' + ctrl.itemName + ' Properties', actionFn: showPropertiesModal },
                { name: 'Edit Properties', title: 'Edit ' + ctrl.itemName + ' Properties', actionFn: showPropertiesModal }
            ];
        }

        function showPropertiesModal(action, item) {
            if (action.name === 'View Properties') {
                ctrl.showModal(Entaxy.MODAL_MODES.VIEW, item);
            } else if (action.name === 'Edit Properties') {
                ctrl.showModal(Entaxy.MODAL_MODES.EDIT, item);
            }
        }

        function populateTable() {
            let items = [];

            let selectedMBean = workspace.getSelectedMBean();
            if (selectedMBean && selectedMBean.isFolder) {
                let currentFolder = selectedMBean;
                if (selectedMBean.text !== ctrl.folderName) {
                    currentFolder = selectedMBean.get(ctrl.folderName);
                }

                if (currentFolder) {
                    let children = currentFolder.children;
                    if (children) {
                        children.forEach((child) => {
                            jolokiaService.getAttributes(child.objectName, [ctrl.statusName])
                            .then((response) => {
                                items.push({name: child.title, status: response[ctrl.statusName], mbeanName: child.objectName});
                            });
                        });

                        ctrl.items = items;
                        ctrl.viewedItems = items;
                        ctrl.toolbarConfig.filterConfig.resultsCount = children.length;
                    }
                }
            }
        }

        ctrl.showModal = function(mode, item) {

            let factories = [];
            let promises = [];

            let factoryFolder = workspace.tree.findDescendant(child => child.title === 'entaxy.runtime.' + ctrl.itemName.toLowerCase());

            if (factoryFolder && factoryFolder.children) {
                factoryFolder.children.forEach((child) => {
                    promises.push(jolokiaService.getAttributes(child.objectName, ['Abstract', 'DisplayName', 'Label', 'Description', 'Deprecated'])
                    .then((result) => {
                        if (!result['Abstract']) {
                            factories.push({
                                name: child.title,
                                displayName: result['DisplayName'] ? result['DisplayName'] : child.title,
                                label: result['Label'],
                                mbeanName: child.objectName,
                                description: result['Description'] ? result['Description'] : 'There is no description for this factory.',
                                additionalInfo: result['Deprecated'] ? ['[DEPRECATED]'] : undefined
                            });
                        }
                    }));
                });

                $q.all(promises).then(() => {
                    if (mode === Entaxy.MODAL_MODES.ADD) {

                        let parentName;
                        if (ctrl.isParentDependent === true) {
                            let selectedMbean = workspace.getSelectedMBean();
                            parentName = selectedMbean.title;
                        }

                        openEntaxyModal({
                            mode: () => mode,
                            itemType: () => ctrl.itemName,
                            parentName: () => parentName,
                            factories: () => factories
                        }, mode);
                    } else {
                        let mbeanName = workspace.getSelectedMBeanName();

                        operationsService
                        .executeOperation(mbeanName, { name: 'get' + ctrl.itemName + 'Config' }, [ item.name ])
                        .then(result => {
                            let itemInfo = JSON.parse(result);

                            let factory = factories.find(factory => factory.name === itemInfo.factoryId);

                            openEntaxyModal({
                                itemName: () => item.name,
                                itemType: () => ctrl.itemName,
                                mode: () => mode,
                                info: () => itemInfo,
                                factories: () => [ factory ]
                            }, mode);
                        }).catch(error => {
                            Core.notification('danger', 'Unable to view configuration of selected item.', 5000);
                            Entaxy.log.error(error);
                        });
                    }
                });
            }
        }

        function openEntaxyModal(resolve, mode) {
            $uibModal.open({
                component: 'entaxyModal',
                resolve: resolve,
                size: 'xl',
                backdrop: 'static'
            })
            .result.then(args => {
                addItem(args, mode)
            },
            reason => {
                if (reason) {
                    Core.notification('danger', reason, 5000);
                }
            });
        }

        function addItem(args, mode) {

            let objectId = args.fields.find((field) => field.name === 'objectId').value;

            let fields = objectId ? args.fields.filter((field) => field.name !== 'objectId') : args.fields;

            let mbeanName = args.factoryId.mbeanName;
            let properties = fields.reduce((obj, cur) => ({ ...obj, [cur.name] : cur.value }), {});
            let instructions = {"@LIFECYCLE":["general"]};

            operationsService.executeOperation(mbeanName, { name: 'createObjectByInstructions' }, [ objectId, 'public', instructions, properties ] )
            .then(result => {
                let tableHtml = Entaxy.createTableFromResponse(result);
                Entaxy.notification('success', tableHtml, 10000);

                if (ctrl.isParentDependent || mode === Entaxy.MODAL_MODES.EDIT) {
                    $scope.$on(Jmx.TreeEvent.Updated, () => populateTable());
                } else {
                    changeLocation(properties[ctrl.objectName] ? properties[ctrl.objectName] : objectId);
                }
            }).catch(error => {
                Core.notification('danger', error, 5000);
                Entaxy.log.error(error);
            });
        }

        function changeLocation(itemName) {
            let currentLocation = $location.search();
            $location.search('nid', currentLocation.nid + '-' + itemName);
        }

        function startItem(action, item) {
            if (item.status !== 'Active') {
                processOperation('start', item.name);
            } else {
                Core.notification('info', ctrl.itemName + ' has been already started');
            }
        }

        function stopItem(action, item) {
            if (item.status !== 'Resolved') {
                processOperation('stop', item.name);
            } else {
                Core.notification('info', ctrl.itemName + ' has been already stopped');
            }
        }

        function uninstallItem(action, item) {
            let operationName = ctrl.isParentDependent ? 'remove' : 'uninstall';
            processOperation(operationName, item.name);
        }

        function processOperation(operationName, selectedItemName) {
            let mbeanName = workspace.getSelectedMBeanName();

            operationsService.executeOperation(mbeanName, { name: operationName + ctrl.itemName }, [ selectedItemName ])
            .then(result => {
                Core.notification('success', result, 3000);
                $scope.$on(Jmx.TreeEvent.Updated, () => populateTable());
            }).catch(error => {
                Core.notification('danger', error, 5000);
                Entaxy.log.error(error);
            });
        }

        ctrl.dismissAlert = () => ctrl.alert = null;

    }
    EntaxyTableController.$inject = ['workspace', '$location', '$scope', '$q', 'jolokia', 'jolokiaService', '$uibModal', 'operationsService'];

})(Entaxy || (Entaxy = {}));
