/*-
 * ~~~~~~licensing~~~~~~
 * entaxy-management-plugin
 * ==========
 * Copyright (C) 2020 - 2023 EmDev LLC
 * ==========
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * ~~~~~~/licensing~~~~~~
 */

var Entaxy;
(function (Entaxy) {
    Entaxy._module.factory('entaxyService', ['$q', 'workspace', 'jolokia', 'jolokiaService', 'operationsService',
        function($q, workspace, jolokia, jolokiaService, operationsService) {
            return {
                makeFormField: function (field, objectId, properties, mode) {
                    return makeFormField(field, objectId, properties, mode);
                },
                getArguments: function (fields, factories) {
                    return getArguments(fields, factories);
                },
                validateFields: function (fields, selectedFactory, itemType) {
                    return validateFields(fields, selectedFactory, itemType);
                }
            };

            function makeFormField(field, objectId, properties, mode) {

                let formField;

                if (mode !== Entaxy.MODAL_MODES.ADD) {

                    if (field.name === 'objectId') {
                        field.value = objectId;
                    }

                    let suitableProperty = properties.find(property => property.name === field.name);

                    if (suitableProperty) {
                        if (field.type.startsWith('entaxy.runtime.')) {
                            field.type = 'string';
                        }

                        if (field.defaultValue && field.defaultValue['@RESOURCE'] && field.defaultValue['@RESOURCE'].format === 'base64') {
                            suitableProperty.value = atob(suitableProperty.value);
                        }

                        field.value = suitableProperty.value;
                    }

                    if (!field.isHidden && field.defaultValue !== undefined && field.value === undefined) {
                        field.displayName = field.displayName ? ('* ' + field.displayName) : ('* ' + field.name);
                    }

                    if (field['@FACADE'] && field['@FACADE'].useFacade) {
                        field.type = field['@FACADE'].type;
                        field.value = field['@FACADE'].value;
                        field.defaultValue = undefined;
                        field.immutable = true;
                    }

                    formField = {
                        label: field.displayName ? field.displayName : field.name,
                        name: field.name,
                        type: field.type,
                        description: field.description,
                        value: field.value ? field.value : field.defaultValue,
                        isRef: field.isRef,
                        isBackRef: field.isBackRef,
                        isHidden: field.isHidden,
                        group: field.group ? field.group : 'general',
                        tag: field.label,
                        readOnly: mode === Entaxy.MODAL_MODES.VIEW ? true : field.immutable,
                        required: mode === Entaxy.MODAL_MODES.EDIT ? field.required : undefined,
                        typeInfo: mode === Entaxy.MODAL_MODES.EDIT ? field['@TYPEINFO'] : undefined,
                        isFromResource: field.defaultValue && field.defaultValue['@RESOURCE'],
                        defaultValue: field.defaultValue,
                        isInternal: field['@INTERNAL']
                    };
                } else {

                    if (field.defaultValue) {
                        if (JSON.stringify(field.defaultValue).indexOf('@CALCULATED') > -1) {
                            field.isCalculated = true;
                            field.placeholder = 'has calculated default value';
                        } else if (field.defaultValue['@RESOURCE']) {
                            field.isFromResource = true;
                        }
                    }

                    formField = {
                        label: field.displayName ? field.displayName : field.name,
                        name: field.name,
                        type: field.type,
                        typeInfo: field['@TYPEINFO'],
                        description: field.description,
                        required: field.required,
                        value: (field.isCalculated || field.isFromResource) ? undefined : field.defaultValue,
                        defaultValue: field.defaultValue,
                        conditional: field.conditional,
                        isRef: field.isRef,
                        isBackRef: field.isBackRef,
                        group: field.group ? field.group : 'general',
                        isHidden: field.isHidden,
                        filter: field.filter ? field.filter.label : undefined,
                        placeholder: field.placeholder,
                        isCalculated: field.isCalculated,
                        isFromResource: field.isFromResource,
                        tag: field.label,
                        uniquenessCheckedProperties: field['@UNIQUE'] ? field['@UNIQUE'].filterProperties : undefined
                    };

                    if (formField.isFromResource) {
                        let resourcePromise = getResource(formField.defaultValue['@RESOURCE']);

                        if (resourcePromise) {
                            resourcePromise
                            .then((response) => { formField.value = response; })
                            .catch((error) => Entaxy.log.error(error));
                        } else {
                            Core.notification('danger', 'Resource service is not found', 5000);
                        }
                    }

                    if (formField.name === '__entaxyContainerId') {
                        let selectedMbean = workspace.getSelectedMBean();

                        jolokiaService.getAttribute(selectedMbean.objectName, 'RuntimeType')
                        .then((runtimeType) => {
                            if (runtimeType && runtimeType.startsWith('entaxy.runtime.')) {
                                formField.value = selectedMbean.title;
                            }
                        })
                        .catch((error) => Entaxy.log.error('There is no runtime type'));
                    }
                }

                return formField;
            }

            function getResource(resourceInfo) {
                let resource = workspace.tree.findDescendant(child => {
                    let childObjectName = child.objectName;

                    if (childObjectName && childObjectName.endsWith('category=resource')) {
                        return child;
                    }
                });

                if (resource) {
                    return operationsService.executeOperation(
                        resource.objectName,
                        { name: 'getResource' },
                        [ resourceInfo.provider + ':' + resourceInfo.location ]
                    );
                }
            }

            function getArguments(fields, factories) {

                let factoryIdName = fields.shift().value;

                return args = {
                    factoryId: factories.find((factory) => factory.name === factoryIdName),
                    fields: fields.map(field => {
                        if (field.isFromResource && field.defaultValue['@RESOURCE'].format === 'base64') {
                            field.value = btoa(field.value);
                        }

                        if (field.isInternal) {
                            field.defaultValue = undefined;
                            field.value = undefined;
                        }

                        return {
                            name: field.name,
                            value: (field.defaultValue !== undefined && field.defaultValue === field.value) ? undefined : field.value
                        };
                    })
                };
            }

            function validateFields(fields, selectedFactory, itemType) {
                let errors = {};

                _.forEach(fields, (field) => {
                    if (field.required && !field.isBackRef && !field.isHidden && !field.isCalculated) {
                        if (field.value === undefined || (typeof field.value === 'string' ? field.value.trim().length === 0 : false)) {
                            errors[field.name] = 'Please fill out this field';
                        }

                        if ((field.name === 'objectId' || field.name === 'systemName') && !errors[field.name]) {
                            if (field.value.trim().length < 3) {
                                errors[field.name] = 'Value must contain at least 3 characters';
                            }
                        }
                    }

                    if (field.uniquenessCheckedProperties && !errors[field.name]) {
                        errors[field.name] = checkUniqueness(field.name, field.value, field.uniquenessCheckedProperties, fields, selectedFactory, itemType);
                    }
                });

                let deferred = $q.defer();

                if (Object.keys(errors).length !== 0) {
                    let errorArrayWithPromises = [];

                    Object.keys(errors).forEach(key => {
                        if (typeof errors[key] !== 'string') {
                            errorArrayWithPromises.push({ name: key, promise: errors[key] });
                        }
                    });

                    if (errorArrayWithPromises.length !== 0) {

                        $q.all(errorArrayWithPromises.map(obj => { return obj.promise; }))
                        .then(result => {
                            for (let i = 0; i < errorArrayWithPromises.length; i++) {
                                if (result[i]) {
                                    errors[errorArrayWithPromises[i].name] = result[i];
                                } else {
                                    delete errors[errorArrayWithPromises[i].name];
                                }
                            }

                            deferred.resolve(errors);
                        });
                    } else {
                        deferred.resolve(errors);
                    }
                } else {
                    deferred.resolve(errors);
                }

                return deferred.promise;
            }

            function checkUniqueness(fieldName, fieldValue, properties, fields, selectedFactory, itemType) {
                let message = 'Value must be unique';
                let isInvalid = false;
                let promises = [];

                properties.unshift(fieldName);
                properties = properties.map(property => {
                    return Entaxy.capitalize(property);
                });
                properties.unshift('RuntimeType');

                let selectedMbean = workspace.getSelectedMBean();

                if (selectedMbean && selectedMbean.isFolder) {
                    let children = Entaxy.getChildrenRecursive(selectedMbean);

                    let childrenMbeanNames = children
                    .map(child => { return child.objectName; })
                    .filter(child => child !== null);

                    childrenMbeanNames.forEach(mbeanName => {
                        promises.push(jolokiaService.getAttributes(mbeanName, properties)
                        .then((response) => {
                            if (response[properties[0]] === 'entaxy.runtime.' + itemType.toLowerCase()
                                    && fieldValue === response[properties[1]]) {
                                for (let i = 2; i < properties.length; i++) {
                                    let checkedField = fields.find(field => field.name === Entaxy.uncapitalize(properties[i]));

                                    if (checkedField) {
                                        if (checkedField.value === response[properties[i]]) {
                                            isInvalid = true;
                                        }
                                    } else {
                                        return jolokiaService.getAttribute(selectedFactory.mbeanName, 'TypeInfo')
                                        .then((typeInfo) => {
                                            let typeInfoDesiredValue = typeInfo[Entaxy.uncapitalize(properties[i])];

                                            if (typeInfoDesiredValue === response[properties[i]]) {
                                                isInvalid = true;
                                            }
                                        });
                                    }
                                }
                            }
                        }));
                    });
                }

                return $q.all(promises).then(() => isInvalid ? message : '');
            }
        }
    ])
})(Entaxy || (Entaxy = {}));
